/*======================================================================
 *                          Copyright (c) 1994
 *                       The Ohio State University
 *               Computer & Information Science Department
 *         Interactive Instructional Computing Facilities (IICF)
 * ======================================================================
 * Permission to use, copy, modify, and  distribute this software and its
 * documentation  for any  purpose  and  without  fee is hereby  granted,
 * provided that the above copyright notice appear in all copies and that
 * both that   the copyright notice  and  warranty  disclaimer appear  in
 * supporting  documentation,   and  that  the  names of   The Ohio State
 * University and  Interactive Instructional Computing Facilities as well
 * as any of   their entities not be  used  in advertising  or  publicity
 * pertaining to distribution of  the software without  specific, written
 * prior permission.
 * 
 * The Ohio State University disclaims all warranties with regard to this
 * software, including   all implied  warranties of   merchantability and
 * fitness.  In no event  shall The Ohio  State University be  liable for
 * any  special,   indirect  or consequential    damages or  any  damages
 * whatsoever resulting from  of use, data or  profits, whether in an
 * action of  contract, negligence or other  tortuous action, arising out
 * of or in connection with the use or performance of this software.
 * =====================================================================
 */
#ifndef TKTREETABLEH
#define TKTREETABLEH

#include <tk.h>
#include <tcl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#ifdef __cplusplus
extern "C" {
#endif

/* tkWinDefault.h defines this, but tkUnixDefault.h doesn't.  */
#ifndef CTL_FONT
#define CTL_FONT "Helvetica -12"
#endif

/*
 * TableItem defaults
 */

#define NO_TABLEITEM_LINEWIDTH          "-1"
#define DEF_TABLEITEM_LINEWIDTH         "1"
#define DEF_TABLEITEM_TEXT              ""
#define DEF_TABLEITEM_DATA              (char *) NULL /*""*/
#define DEF_TABLEITEM_BITMAP            (char *) NULL
#define DEF_TABLEITEM_IMAGE             (char *) NULL
#define DEF_TABLEITEM_IDLECOMMAND       (char *) NULL
#define DEF_HIDDEN_COMMAND              ""
#define DEF_TABLEITEM_FONT              ""
#define DEF_TABLEITEM_TEXTFOREGROUND    ""
#define DEF_TABLEITEM_BITMAPFOREGROUND  ""
#define DEF_TABLEITEM_BITMAPBACKGROUND  NORMAL_BG
#define DEF_TABLEITEM_BITMAPSELECTFOREGROUND ""
#define DEF_TABLEITEM_BITMAPSELECTBACKGROUND ""
#define DEF_TABLEITEM_LINEFOREGROUND    ""
#define DEF_TABLEITEM_INDENT            "0"
#define DEF_TABLEITEM_PARENT            "-1"
#define DEF_TABLEITEM_SUCCNUM           "0"
#define DEF_TABLEITEM_SEENNUM           "0"
#define DEF_TABLEITEM_TABFREESPACE      "6"
#define DEF_TABLEITEM_FLAGS             "0"

/*
 * TreeTable defaults
 */

#define NO_TREETABLE_LINEWIDTH          "-1"
#define DEF_TREETABLE_LINEWIDTH         "1"
#define DEF_TREETABLE_BG_COLOR          DEF_TEXT_BG_COLOR
#define DEF_TREETABLE_BG_MONO           DEF_TEXT_BG_MONO
#define DEF_TREETABLE_BORDER_WIDTH      "2"
#define DEF_TREETABLE_HIGHLIGHT_WIDTH   "2"
#define DEF_TREETABLE_CURSOR            ""
#define DEF_TREETABLE_EXPORT_SELECTION  "0"
#define DEF_TREETABLE_TEXTFG            DEF_TEXT_FG
#define DEF_TREETABLE_BITMAPFG          BLACK
#define DEF_TREETABLE_BITMAPBG          NORMAL_BG
#define DEF_TREETABLE_LINEFG            BLACK
#define DEF_TREETABLE_SPLITLINEFG       "#c0c0c0"
#ifdef _TKWINDEFAULT
#define DEF_TREETABLE_FONT              DEF_TEXT_FONT
#else
/* For compatibility.  */
#define DEF_TREETABLE_FONT              CTL_FONT
#endif
#define DEF_TREETABLE_FG                DEF_TEXT_FG
#define DEF_TREETABLE_GEOMETRY          "20x10"
#define DEF_TREETABLE_WIDTH             "20"
#define DEF_TREETABLE_HEIGHT            "10"
#define DEF_TREETABLE_RELIEF            "flat"
#define DEF_TREETABLE_SELECT_COLOR      DEF_TEXT_SELECT_COLOR
#define DEF_TREETABLE_SELECT_MONO       DEF_TEXT_SELECT_MONO
#define DEF_TREETABLE_INDENTWIDTH       "10"
#define DEF_TREETABLE_BITMAPSPACE       "7"
#define DEF_TREETABLE_SELECT_BD         "1"
#define DEF_TREETABLE_SELECT_FG_COLOR   DEF_TEXT_SELECT_FG_COLOR
#define DEF_TREETABLE_SELECT_FG_MONO    DEF_TEXT_SELECT_FG_MONO
#define DEF_TREETABLE_BITMAPSELECTBACKGROUND ""
#define DEF_TREETABLE_BITMAPSELECTFOREGROUND ""

#define DEF_TREETABLE_SET_GRID          "0"
#define DEF_TREETABLE_BESTFIT           "0"
#define DEF_TREETABLE_AUTOFIT           "0"
#define DEF_TREETABLE_SPLITLINES        "1"
#define DEF_TREETABLE_TRUNCATE          "1"
#define DEF_TREETABLE_TAKE_FOCUS        (char *) NULL
#define DEF_TREETABLE_SCROLL_COMMAND    ""
#define DEF_TREETABLE_TABS              (char *) NULL
#define DEF_TREETABLE_JUSTIFY           (char *) NULL
#define DEF_TREETABLE_DEFTABS           (char *) NULL
#define DEF_TREETABLE_SELECTMODE	"browse"
#define DEF_TREETABLE_FILL_SELECTION    "0"
#define DEF_TREETABLE_SORTED_INSERTION  "0"
#define DEF_TREETABLE_SORT_NOCASE       "0"
#define DEF_TREETABLE_SORT_COLNUM       "-1"
#define DEF_TREETABLE_USE_ACCELERATOR   "0"

#define ITEM_HIDDEN_SUBTREE		0x0001
#define ITEM_SELECTED			0x0002

#define MULTI				0
#define SINGLE				1

#undef USE_PARENT_CACHE
#undef USE_CACHE
#define MAX_CACHED			5

#undef USE_BITMAP_COLORS
#define USE_PATHFINDER

/*
 * This is the data kept for each line in the treetable.
 */
typedef struct TableItem {
    short bitmapHeight;		/* Height of the bitmap */
    short bitmapWidth;		/* Width of the bitmap */
    
    short indent;			/* Indentation level for the line */
    short lineWidth;		/* Width of the line joining this item to
				 * it's ancestor, measured in pixels. */
    short flags;
    short fontHeight;		/* Height of the font -- includes ascent
				 * and descent */
				 
    Pixmap bitmap;		/* Bitmap to display for the line */
    
    int textLength;		/* number of chars in text label */
    int parent;			/* This is the index for the parent 
				 * of this node. */
    int succNum;		/* Number of all sons, in sub trees too */
    int seenNum;		/* seen item number of sub tree */
    GC textGC;			/* Graphics context to use for drawing the
				 * text to the screen */
    GC bitmapGC;		/* Graphics context to use for draing the
				 * bitmap to the screen */
    
    char *text;			/* Pointer to the text label string.  If 
				 * NULL, then there is no label for
				 * the line */
    char *data;			/* Pointer to client data to store
				 * private informations, like flags */
    
    Tk_Image image;		/* we can load images */
    char *imageString;		/* image name */
    
    Tk_Font fontPtr;		/* Font pointer to use for text */
    XColor *textFgColor;	/* Color to write text in */
    
#ifdef USE_BITMAP_COLORS
    GC bitmapSelectGC;           /* GC to use for bitmap when item selected */
    GC lineGC;			 /* Graphics context for the line */
    XColor *bitmapFgColor;	 /* foreground color for bitmap */
    XColor *bitmapBgColor;	 /* background color for the bitmap */
    XColor *bitmapSelectFgColor; /* Bitmap Fg color when item selected */
    XColor *bitmapSelectBgColor; /* Bitmap Bg color when item is selected */ 
    XColor *lineFgColor;	 /* line color */
#endif

    struct TableItem *succPtr;	 /* Sons of the tree */
    struct TableItem *parentPtr; /* Pointer to the TableItem which is our
				  * ancestor, or NULL if we don't have a
				  * parent. */
    struct TableItem *nextPtr;	 /* Pointer to the next item in the table */
} TableItem;

#ifdef USE_PATHFINDER
/*
 * This pathfinder is used to accelerate the display routine,
 * exactly when a single list has too many entries (> 20t).
 * The Pathfinder saves 'stop stations' every a number of
 * items in the root list (every 1t as example).
 * So, when an item is searched, numbered 25t) we will find
 * an previous item on the position 20t (in 20 Steps ansted of 20t steps)
 * and we continue normal searching in the left 500 entries, this costs
 * only time of 20+500 = 520 ansted of 20t.
 *
 * The path is built by the treetable routines, as example
 * display procedure, insertion, deleting ans so on.
 * The Path is usually refreshed by deleting or adding an item.
 */
#define PATHFINDER_STEP 1024
typedef struct PathFinder_t {
    TableItem* itemPtr;
    int itemPos;
    struct PathFinder_t * next;
} PathFinder_t;
#endif

/*
 * Data structure used for each widget of type TreeTable in
 * the interpreter.  (This a bastardization of tkListbox.c)
 */
typedef struct {
    Tk_Window tkwin;            /* Window that embodies the TreeTable. NULL
				 * means that the window has been destroyed
				 * but the data structures haven't yet been
				 * cleaned up. */
    Display *display;           /* Display containing the widget.  Used, among
				 * other things, so that resources can be freed
				 * even after tkwin has gone away. */
    Tcl_Interp *interp;         /* Interpreter associated with this widget */
    Tcl_Command widgetCmd;	/* Token for listbox's widget command. */
    int numItems;               /* Size of table for the widget */
    int numViewed;              /* number of viewed items, not hidden sub trees */
    TableItem *itemPtr;         /* Pointer to table items -- NULL if there 
                                 * are none. */
#ifdef USE_PATHFINDER
    int UseAccelerator;		/* use pathfinder */
    PathFinder_t *pathFinder;	/* To find elements in the middle of the
    				 * item list of the treetable, when we
				 * have alot of items (more than 5000) as
				 * root elements
				 * This is requiered for fast displaying the
				 * list.
				 */
#endif

#ifdef USE_CACHE
    TableItem *cachedItem[MAX_CACHED];	/* Heuristec cache */
    int cachedPos[MAX_CACHED];
    int cacheIndex;			/* points to latest cache position */
#endif
#ifdef USE_PARENT_CACHE
    int cachedParent;
    TableItem *cachedParentPtr;
#endif
    
    /*
     * Information used when displaying the widget :
     */
    
    Tk_3DBorder normalBorder;   /* Used for drawing border around the whole
				 * window, plus used for background */
    int borderWidth;            /* Width of 3-D border around window */
    int relief;                 /* 3-D effect : TK_RELIEF_RAISED, etc... */
    int highlightWidth;         /* Width in pixels of highlight to draw
				 * around widget when it has the focus.
				 * <= 0 means don't draw a highlight */
    XColor *highlightBgColor; 
                                /* Color for drawing traversal highlight
				 * area when highlight is off. */
    XColor *highlightColor;     /* Color for drawing traversal highlight. */
    int inset;                  /* Total width of all borders, including
				 * traversal highlight and 3-D border.
				 * Indicates how much interior stuff must
				 * be offset from outside edges to leave
				 * room for borders. */
    Tk_Font defFontPtr;		/* Information about text font, or NULL. */
    int defFontHeight;          /* Height of the default font -- includes
				 * ascent and descent */
    Tk_3DBorder selBorder;      /* Borders and backgrounds for selected
				 * items. */
    int selBorderWidth;         /* Width of border around selection */
    int defLineWidth;           /* Default width of child-parent lines */
    XColor *defTextFgColor;     /* Default foreground color for text */
    XColor *defBitmapFgColor;   /* Default foreground for bitmaps */
    XColor *defBitmapBgColor;   /* Default background for bitmaps */
    XColor *defBitmapSelectFgColor;
    XColor *defBitmapSelectBgColor;
    XColor *defLineFgColor;     /* Default foreground color for lines */
    XColor *selFgColorPtr;      /* Foreground color for select items */
    GC selectGC;                /* For drawing selected items */
    GC defTextGC;               /* Default graphics context for text */
    GC defBitmapGC;             /* Default graphics context for bitmaps */
    GC defBitmapSelectGC;       /* Default graphics context for bitmaps when
				 * selection is active for the line */
    GC defLineGC;               /* Default graphics context for lines */
    char *geometry;             /* Desired geometry for window. Malloc'ed */
    int width, height;          /* another possibility to configure treetable size */
    int lineHeight;             /* Number of pixels allocated for each line 
				 * in the display (tallest out of fonts
				 * or bitmaps */
    int bitmapSpace;            /* Width of empty space between the bitmap
				 * of a line and the text associated with it,
				 * measured in pixels. */
    int indentWidth;            /* Indentation level expressed as number 
				 * of pixels from the left side of the
				 * widget. */
    int topIndex;               /* Index of topmost item in the window */
    int numLines;               /* Number of lines(items) that fit
				 * in window at one time */
    int setGrid;                /* Non-zero means pass gridding information
				 * to window manager */
    
    Pixmap hiddenBitmap;	/* Bitmap to display when sub tree is hidden */
    Tk_Image hiddenImage;	/* Image to display when sub tree is hidden */
    char *hiddenImageString;
    int hiddenWidth;		/* Bitmap to display when sub tree is hidden */
    int hiddenHeight;		/* Bitmap to display when sub tree is hidden */
#if 0
    char * hiddencommand;	/* command to execute when a sub tree is toggled */
#endif
    char * idlecommand;	/* command to execute */
  
    /*
     * Information to support horizontal scrolling 
     */
    int maxWidth;               /* Width (in pixels) of widest line in the
				 * widget. */
    int xScrollUnit;            /* Number of pixels in one "unit" for
				 * horizontal scrolling (window scrolls
				 * horizontally in increments of this size).
				 * This is an average character size. */
    int xOffset;                /* The left edge of each line in the treetable
				 * is offset to the left by this many
				 * pixels (0 means no offset, positive 
				 * means there is an offset). */

    /*
     * Information for scanning:
     */

    int scanMarkX;              /* X-position at which scan started (e.g.
                                 * button was pressed here). */
    int scanMarkY;              /* Y-position at which scan started (e.g.
                                 * button was pressed here). */
    int scanMarkXOffset;        /* Value of "xOffset" field when scan
                                 * started. */
    int scanMarkYIndex;         /* Index of line that was at top of window
                                 * when scan started. */


    
    /*
     * Information about what's selected, if any
     */
    TableItem *lastSelected;
    int numSelected;            /* selected number of items */
    int selectFirst;            /* Index of first selected item (-1 means
				 * nothing selected). */
    int selectAnchor;           /* Line to which the selection is anchored */
    int exportSelection;        /* Non-zero means tie internal value
				 * to X selection. */
    int fillSelection;          /* Non-zero means the selection box will
				 * expand to max. width of window and not
				 * only text region */
    int active;                 /* Index of "active" element (the one that
				 * has been selected by keyboard traversal)
				 * -1 means none. */
    int   selectMode;           /* MULTI or SINGLE */
    char* selectModeStr;        /* multi or single */

    /*
     * we support tabs too
     */
    int * defTabs;			/* converted defualt tab stop list */
    char *defTabsList;		/* Tcl List with defualt tab stops */
    
    char *tabsList;		/* Tcl List with tab stops */
    int * tabs;			/* converted tab list */
    int tabsNum;		/* number of tabs */
    int tabsMinSpace;		/* Min. Space between tow tab stops */
    int *tabsHidden;		/* tabs that are hidden */
    
    /*
     * We support right-justify too
     */
    char *justify;
    int *tabsJustify;
    
    /*
     * Miscellaneous information ;
     */
    Cursor cursor;              /* X cursor for the window, or None. */
    char *takeFocus;            /* Value of -takefocus option;  not used in
				 * the C code, but used by keyboard traversal
				 * scripts.  Malloc'ed, but may be NULL */
    char *yScrollCmd;           /* Command prefix for communicating with
				 * vertical scrollbar.  NULL means no
				 * command to issue.  Malloc'ed. */
    char *xScrollCmd;           /* Command prefix for communicating with
				 * horizontal scrollbar.  NULL means no
				 * command to issue.  Malloc'ed */
    int flags;                  /* Various flag bits:  see below for
				 * definitions. */
    int sortedInsertion;        /* if this flag is seted, the items will be inserted
    				 * on the correct place */
    int sortNoCase;		/* to ignore upper and lower case by comparing
    				 * to place items correctly */
    int sortColumn;		/* insert items sorted using the strin in the
    				 * tab stop of the given number, default "-1" */
    int pressX, pressY;
    
    int BestFit;		/* Best fit for tab stops */
    int AutoFit;		/* Change Tab Stop Size of the current
    				 * displayed items so that we have best
				 * fit, this flag overrites 'BestFit' */
    int Truncate;	/* Truncate the rest of string */
    int SplitLines;
    
    XColor *defSplitLineFgColor; /* Default foreground color for split lines */
    GC defSplitLineGC;           /* Default graphics context for split lines */
} TreeTable;

/*
 * Flag bits for TreeTable
 * REDRAW_PENDING :            Non-zero means a DoWhenIdle handler 
 *                             has already been queued to redraw
 *                             this window.
 * UPDATE_V_SCROLLBAR :        Non-zero means verticle scrollbar needs
 *                             to be updated.
 * UPDATE_H_SCROLLBAR:         Non-zero means horizontal scrollbar needs
 *                             to be updated.
 * GOT_FOCUS:                  Non-zero means this widget currently
 *                             has the input focus.
 */

#define REDRAW_PENDING        1
#define UPDATE_V_SCROLLBAR    2
#define UPDATE_H_SCROLLBAR    4
#define GOT_FOCUS             8

/*
 * Function declaration
 */

int create_treetable_command _ANSI_ARGS_((Tcl_Interp *interp));

#ifdef __cplusplus
}
#endif

#endif
